@@ -1,5 +1,5 @@
 /* libinfinity - a GObject-based infinote implementation
- * Copyright (C) 2007-2011 Armin Burgmeier <armin@arbur.net>
+ * Copyright (C) 2007-2014 Armin Burgmeier <armin@arbur.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -17,6 +17,28 @@
  * MA 02110-1301, USA.
  */
 
+/**
+ * SECTION:inf-standalone-io
+ * @title: InfStandaloneIo
+ * @short_description: Simple event loop implementation
+ * @include: libinfinity/common/inf-standalone-io.h
+ * @see_also: #InfIo
+ * @stability: Unstable
+ *
+ * #InfStandaloneIo is a simple implementation of the #InfIo interface. It
+ * implements a basic application event loop with support for listening on
+ * sockets, scheduling timeouts and inter-thread notifications. The class
+ * is fully thread-safe.
+ *
+ * This class can be perfectly used for all functions in libinfinity that
+ * require a #InfIo object to wait for events. If, on top of that more
+ * functionality is required, or the main loop needs to be integrated with
+ * another library such as a UI toolkit, a custom class should be created
+ * instead which implements the #InfIo interface. For the GTK+ toolkit, there
+ * is #InfGtkIo in the libinfgtk library, to integrate with the Glib main
+ * loop.
+ */
+
 #include <libinfinity/common/inf-standalone-io.h>
 #include <libinfinity/common/inf-io.h>
 
@@ -860,6 +882,22 @@ inf_standalone_io_io_remove_watch(InfIo* io,
   watch_iter = inf_standalone_io_find_watch(INF_STANDALONE_IO(io), watch);
   if(watch_iter != NULL)
   {
+#ifdef G_OS_WIN32
+    if(WSAEventSelect(*watch->socket, *watch->event, 0) == SOCKET_ERROR)
+    {
+      error_message = g_win32_error_message(WSAGetLastError());
+      g_warning("WSAEventSelect() failed: %s", error_message);
+      g_free(error_message);
+    }
+
+    if(WSACloseEvent(*watch->event) == FALSE)
+    {
+      error_message = g_win32_error_message(WSAGetLastError());
+      g_warning("WSACloseEvent() failed: %s", error_message);
+      g_free(error_message);
+    }
+#endif
+
     /* TODO: If we are currently polling we should not modify the fds array
      * array but do this after wakeup directly after the poll call. */
     if(watch->executing)
@@ -878,24 +916,8 @@ inf_standalone_io_io_remove_watch(InfIo* io,
       g_slice_free(InfIoWatch, watch);
     }
 
-#ifdef G_OS_WIN32
-    if(WSAEventSelect(*watch->socket, *watch->event, 0) == SOCKET_ERROR)
-    {
-      error_message = g_win32_error_message(WSAGetLastError());
-      g_warning("WSAEventSelect() failed: %s", error_message);
-      g_free(error_message);
-    }
-
-    if(WSACloseEvent(*watch->event) == FALSE)
-    {
-      error_message = g_win32_error_message(WSAGetLastError());
-      g_warning("WSACloseEvent() failed: %s", error_message);
-      g_free(error_message);
-    }
-#endif
-
     /* Remove watch by replacing it by the last pollfd/watch */
-    index = 1 + watch_iter - priv->watches;
+    index = 1 + (watch_iter - priv->watches);
     if(index != priv->fd_size - 1)
     {
       memcpy(
