@@ -1,5 +1,5 @@
 /* libinfinity - a GObject-based infinote implementation
- * Copyright (C) 2007-2011 Armin Burgmeier <armin@arbur.net>
+ * Copyright (C) 2007-2014 Armin Burgmeier <armin@arbur.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -17,10 +17,36 @@
  * MA 02110-1301, USA.
  */
 
+/**
+ * SECTION:inf-tcp-connection
+ * @title: InfTcpConnection
+ * @short_description: Represents a TCP-based connection to a remote host
+ * @include: libinfinity/common/inf-tcp-connection.h
+ * @stability: Unstable
+ *
+ * #InfTcpConnection represents a TCP connection to a remove host. It is a
+ * wrapper around a native socket object and integrates into the main loop
+ * provided by #InfIo. An arbitrary amount of data can be sent with the
+ * object, extra data will be buffered and automatically transmitted once
+ * kernel space becomes available.
+ *
+ * The TCP connection properties should be set and then
+ * inf_tcp_connection_open() be called to open a connection. If the
+ * #InfTcpConnection:resolver property is set, then
+ * #InfTcpConnection:remote-address and #InfTcpConnection:remote-port are
+ * ignored, and the hostname as configured in the resolver will be resolved.
+ * When the hostname has been resolved and a connection has been made, the
+ * #InfTcpConnection:remote-address and #InfTcpConnection:remote-port
+ * properties are updated to reflect the address actually connected to.
+ **/
+
 #include <libinfinity/common/inf-tcp-connection.h>
 #include <libinfinity/common/inf-tcp-connection-private.h>
+#include <libinfinity/common/inf-name-resolver.h>
 #include <libinfinity/common/inf-ip-address.h>
 #include <libinfinity/common/inf-io.h>
+#include <libinfinity/common/inf-native-socket.h>
+#include <libinfinity/inf-signals.h>
 #include <libinfinity/inf-marshal.h>
 #include <libinfinity/inf-i18n.h>
 
@@ -43,34 +69,15 @@
 # include <ws2tcpip.h>
 #endif
 
-#ifdef G_OS_WIN32
-# define INF_TCP_CONNECTION_SENDRECV_FLAGS 0
-# define INF_TCP_CONNECTION_LAST_ERROR     WSAGetLastError()
-# define INF_TCP_CONNECTION_EINTR          WSAEINTR
-# define INF_TCP_CONNECTION_EAGAIN         WSAEWOULDBLOCK
-/* This is not a typo here. On Windows, connect() returns WSAEWOULDBLOCK on
- * a non-blocking socket. */
-# define INF_TCP_CONNECTION_EINPROGRESS    WSAEWOULDBLOCK
-#else
-# ifdef HAVE_MSG_NOSIGNAL
-#  define INF_TCP_CONNECTION_SENDRECV_FLAGS MSG_NOSIGNAL
-# else
-#  define INF_TCP_CONNECTION_SENDRECV_FLAGS 0
-# endif
-# define INF_TCP_CONNECTION_LAST_ERROR     errno
-# define INF_TCP_CONNECTION_EINTR          EINTR
-# define INF_TCP_CONNECTION_EAGAIN         EAGAIN
-# define INF_TCP_CONNECTION_EINPROGRESS    EINPROGRESS
-# define closesocket(s) close(s)
-# define INVALID_SOCKET -1
-#endif
-
 typedef struct _InfTcpConnectionPrivate InfTcpConnectionPrivate;
 struct _InfTcpConnectionPrivate {
   InfIo* io;
   InfIoEvent events;
   InfIoWatch* watch;
 
+  InfNameResolver* resolver;
+  guint resolver_index;
+
   InfTcpConnectionStatus status;
   InfNativeSocket socket;
 
@@ -88,6 +95,7 @@ enum {
   PROP_0,
 
   PROP_IO,
+  PROP_RESOLVER,
 
   PROP_STATUS,
 
@@ -113,40 +121,6 @@ enum {
 static GObjectClass* parent_class;
 static guint tcp_connection_signals[LAST_SIGNAL];
 
-static GQuark
-inf_tcp_connection_error_quark(void)
-{
-  return g_quark_from_static_string("INF_TCP_CONNECTION_ERROR");
-}
-
-static void
-inf_tcp_connection_make_system_error(int code,
-                                     GError** error)
-{
-#ifdef G_OS_WIN32
-  gchar* error_message;
-  error_message = g_win32_error_message(code);
-
-  g_set_error(
-    error,
-    inf_tcp_connection_error_quark(),
-    code,
-    "%s",
-    error_message
-  );
-
-  g_free(error_message);
-#else
-  g_set_error(
-    error,
-    inf_tcp_connection_error_quark(),
-    code,
-    "%s",
-    strerror(code)
-  );
-#endif
-}
-
 static gboolean
 inf_tcp_connection_addr_info(InfNativeSocket socket,
                              gboolean local,
@@ -172,8 +146,8 @@ inf_tcp_connection_addr_info(InfNativeSocket socket,
 
   if(res == -1)
   {
-    code = INF_TCP_CONNECTION_LAST_ERROR;
-    inf_tcp_connection_make_system_error(code, error);
+    code = INF_NATIVE_SOCKET_LAST_ERROR;
+    inf_native_socket_make_error(code, error);
     return FALSE;
   }
 
@@ -206,7 +180,7 @@ inf_tcp_connection_system_error(InfTcpConnection* connection,
   GError* error;
   error = NULL;
 
-  inf_tcp_connection_make_system_error(code, &error);
+  inf_native_socket_make_error(code, &error);
 
   g_signal_emit(
     G_OBJECT(connection),
@@ -218,69 +192,12 @@ inf_tcp_connection_system_error(InfTcpConnection* connection,
   g_error_free(error);
 }
 
-static gboolean
-inf_tcp_connection_send_real(InfTcpConnection* connection,
-                             gconstpointer data,
-                             guint* len)
-{
-  InfTcpConnectionPrivate* priv;
-  gconstpointer send_data;
-  guint send_len;
-  int errcode;
-  ssize_t result;
-
-  priv = INF_TCP_CONNECTION_PRIVATE(connection);
-  g_assert(priv->status == INF_TCP_CONNECTION_CONNECTED);
-
-  g_assert(data != NULL);
-  g_assert(len != NULL);
-
-  send_data = data;
-  send_len = *len;
-
-  do
-  {
-    result = send(
-      priv->socket,
-      send_data,
-      send_len,
-      INF_TCP_CONNECTION_SENDRECV_FLAGS
-    );
-
-    /* Preserve error code so that it is not modified by future calls */
-    errcode = INF_TCP_CONNECTION_LAST_ERROR;
-
-    if(result < 0 &&
-       errcode != INF_TCP_CONNECTION_EINTR &&
-       errcode != INF_TCP_CONNECTION_EAGAIN)
-    {
-      inf_tcp_connection_system_error(connection, errcode);
-      return FALSE;
-    }
-    else if(result == 0)
-    {
-      inf_tcp_connection_close(connection);
-      return FALSE;
-    }
-    else if(result > 0)
-    {
-      send_data = (const char*)send_data + result;
-      send_len -= result;
-    }
-  } while( (send_len > 0) &&
-           (result > 0 || errcode == INF_TCP_CONNECTION_EINTR) &&
-           (priv->socket != INVALID_SOCKET) );
-
-  *len -= send_len;
-  return TRUE;
-}
-
-/* Required by inf_tcp_connection_connected */
 static void
 inf_tcp_connection_io(InfNativeSocket* socket,
                       InfIoEvent events,
                       gpointer user_data);
 
+
 static void
 inf_tcp_connection_connected(InfTcpConnection* connection)
 {
@@ -310,12 +227,364 @@ inf_tcp_connection_connected(InfTcpConnection* connection)
   }
 
   g_object_freeze_notify(G_OBJECT(connection));
+
+  /* Update adresses from resolver */
+  if(priv->resolver != NULL)
+  {
+    if(priv->remote_address != NULL)
+      inf_ip_address_free(priv->remote_address);
+
+    priv->remote_address = inf_ip_address_copy(
+      inf_name_resolver_get_address(priv->resolver, priv->resolver_index)
+    );
+
+    priv->remote_port =
+      inf_name_resolver_get_port(priv->resolver, priv->resolver_index);
+
+    g_object_notify(G_OBJECT(connection), "remote-address");
+    g_object_notify(G_OBJECT(connection), "remote-port");
+
+    priv->resolver_index = 0;
+  }
+
   g_object_notify(G_OBJECT(connection), "status");
   g_object_notify(G_OBJECT(connection), "local-address");
   g_object_notify(G_OBJECT(connection), "local-port");
   g_object_thaw_notify(G_OBJECT(connection));
 }
 
+static gboolean
+inf_tcp_connection_open_with_resolver(InfTcpConnection* connection,
+                                      GError** error);
+
+/* Handles when an error occurred during connection. Returns FALSE when the
+ * error was fatal. In this case, it has already emitted the "error" signal.
+ * Returns TRUE, if another connection attempt is made. */
+static gboolean
+inf_tcp_connection_connection_error(InfTcpConnection* connection,
+                                    const GError* error)
+{
+  InfTcpConnectionPrivate* priv;
+
+  priv = INF_TCP_CONNECTION_PRIVATE(connection);
+
+  if(priv->socket != INVALID_SOCKET)
+  {
+    closesocket(priv->socket);
+    priv->socket = INVALID_SOCKET;
+  }
+
+  if(priv->watch != NULL)
+  {
+    priv->events = 0;
+
+    inf_io_remove_watch(priv->io, priv->watch);
+    priv->watch = NULL;
+  }
+
+  if(priv->resolver != NULL)
+  {
+    /* Try next address, if there is one */
+    if(priv->resolver_index < inf_name_resolver_get_n_addresses(priv->resolver))
+    {
+      ++priv->resolver_index;
+      if(inf_tcp_connection_open_with_resolver(connection, NULL) == TRUE)
+        return TRUE;
+    }
+
+    /* No new addresses available */
+    priv->resolver_index = 0;
+  }
+
+  g_signal_emit(
+    G_OBJECT(connection),
+    tcp_connection_signals[ERROR_],
+    0,
+    error
+  );
+
+  return FALSE;
+}
+
+static gboolean
+inf_tcp_connection_open_real(InfTcpConnection* connection,
+                             const InfIpAddress* address,
+                             guint port,
+                             GError** error)
+{
+  InfTcpConnectionPrivate* priv;
+
+#ifdef G_OS_WIN32
+  u_long argp;
+#endif
+
+  union {
+    struct sockaddr_in in;
+    struct sockaddr_in6 in6;
+  } native_address;
+
+  struct sockaddr* addr;
+  socklen_t addrlen;
+  int result;
+  int errcode;
+  GError* local_error;
+
+  priv = INF_TCP_CONNECTION_PRIVATE(connection);
+
+  g_assert(priv->status == INF_TCP_CONNECTION_CLOSED ||
+           priv->status == INF_TCP_CONNECTION_CONNECTING);
+
+  /* Close previous socket */
+  if(priv->socket != INVALID_SOCKET)
+    closesocket(priv->socket);
+
+  switch(inf_ip_address_get_family(address))
+  {
+  case INF_IP_ADDRESS_IPV4:
+    priv->socket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
+    addr = (struct sockaddr*)&native_address.in;
+    addrlen = sizeof(struct sockaddr_in);
+
+    memcpy(
+      &native_address.in.sin_addr,
+      inf_ip_address_get_raw(address),
+      sizeof(struct in_addr)
+    );
+
+    native_address.in.sin_family = AF_INET;
+    native_address.in.sin_port = htons(port);
+
+    break;
+  case INF_IP_ADDRESS_IPV6:
+    priv->socket = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
+    addr = (struct sockaddr*)&native_address.in6;
+    addrlen = sizeof(struct sockaddr_in6);
+
+    memcpy(
+      &native_address.in6.sin6_addr,
+      inf_ip_address_get_raw(address),
+      sizeof(struct in6_addr)
+    );
+
+    native_address.in6.sin6_family = AF_INET6;
+    native_address.in6.sin6_port = htons(port);
+    native_address.in6.sin6_flowinfo = 0;
+    native_address.in6.sin6_scope_id = priv->device_index;
+
+    break;
+  default:
+    g_assert_not_reached();
+    break;
+  }
+
+  if(priv->socket == INVALID_SOCKET)
+  {
+    inf_native_socket_make_error(INF_NATIVE_SOCKET_LAST_ERROR, error);
+    return FALSE;
+  }
+
+  /* Set socket non-blocking */
+#ifndef G_OS_WIN32
+  result = fcntl(priv->socket, F_GETFL);
+  if(result == INVALID_SOCKET)
+  {
+    errcode = INF_NATIVE_SOCKET_LAST_ERROR;
+    inf_native_socket_make_error(errcode, error);
+
+    closesocket(priv->socket);
+    priv->socket = INVALID_SOCKET;
+    return FALSE;
+  }
+
+  if(fcntl(priv->socket, F_SETFL, result | O_NONBLOCK) == -1)
+  {
+    errcode = INF_NATIVE_SOCKET_LAST_ERROR;
+    inf_native_socket_make_error(errcode, error);
+
+    closesocket(priv->socket);
+    priv->socket = INVALID_SOCKET;
+    return FALSE;
+  }
+#else
+  argp = 1;
+  if(ioctlsocket(priv->socket, FIONBIO, &argp) != 0)
+  {
+    errcode = INF_NATIVE_SOCKET_LAST_ERROR;
+    inf_native_socket_make_error(errcode, error);
+
+    closesocket(priv->socket);
+    priv->socket = INVALID_SOCKET;
+    return FALSE;
+  }
+#endif
+
+  /* Connect */
+  do
+  {
+    result = connect(priv->socket, addr, addrlen);
+    errcode = INF_NATIVE_SOCKET_LAST_ERROR;
+    if(result == -1 &&
+       errcode != INF_NATIVE_SOCKET_EINTR &&
+       errcode != INF_NATIVE_SOCKET_EINPROGRESS)
+    {
+      local_error = NULL;
+      inf_native_socket_make_error(errcode, &local_error);
+      if(inf_tcp_connection_connection_error(connection, local_error) == TRUE)
+      {
+        /* In this case, we could recover from the error by connecting to a
+         * different address. */
+        g_error_free(local_error);
+        return TRUE;
+      }
+
+      g_propagate_error(error, local_error);
+      return FALSE;
+    }
+  } while(result == -1 && errcode != INF_NATIVE_SOCKET_EINPROGRESS);
+
+  if(result == 0)
+  {
+    /* Connection fully established */
+    inf_tcp_connection_connected(connection);
+  }
+  else
+  {
+    g_assert(priv->watch == NULL);
+
+    /* Connection establishment in progress */
+    priv->events = INF_IO_OUTGOING | INF_IO_ERROR;
+
+    priv->watch = inf_io_add_watch(
+      priv->io,
+      &priv->socket,
+      priv->events,
+      inf_tcp_connection_io,
+      connection,
+      NULL
+    );
+
+    if(priv->status != INF_TCP_CONNECTION_CONNECTING)
+    {
+      priv->status = INF_TCP_CONNECTION_CONNECTING;
+      g_object_notify(G_OBJECT(connection), "status");
+    }
+  }
+
+  return TRUE;
+}
+
+static gboolean
+inf_tcp_connection_open_with_resolver(InfTcpConnection* connection,
+                                      GError** error)
+{
+  InfTcpConnectionPrivate* priv;
+  GError* local_error;
+  gboolean success;
+
+  priv = INF_TCP_CONNECTION_PRIVATE(connection);
+
+  g_assert(priv->status == INF_TCP_CONNECTION_CLOSED ||
+           priv->status == INF_TCP_CONNECTION_CONNECTING);
+
+  if(inf_name_resolver_finished(priv->resolver))
+  {
+    if(priv->resolver_index < 
+       inf_name_resolver_get_n_addresses(priv->resolver))
+    {
+      return inf_tcp_connection_open_real(
+        connection,
+        inf_name_resolver_get_address(priv->resolver, priv->resolver_index),
+        inf_name_resolver_get_port(priv->resolver, priv->resolver_index),
+        error
+      );
+    }
+
+    /* We need to look up more addresses */
+    g_object_freeze_notify(G_OBJECT(connection));
+    if(priv->status != INF_TCP_CONNECTION_CONNECTING)
+    {
+      priv->status = INF_TCP_CONNECTION_CONNECTING;
+      g_object_notify(G_OBJECT(connection), "status");
+    }
+
+    local_error = NULL;
+    if(priv->resolver_index == 0)
+      success = inf_name_resolver_start(priv->resolver, &local_error);
+    else
+      success = inf_name_resolver_lookup_backup(priv->resolver, &local_error);
+
+    if(local_error != NULL)
+    {
+      inf_tcp_connection_connection_error(connection, local_error);
+      g_propagate_error(error, local_error);
+    }
+
+    g_object_thaw_notify(G_OBJECT(connection));
+    return success;
+  }
+
+  /* The resolver is currently doing something. Wait until it finishes, and
+   * then try again. */
+  return TRUE;
+}
+
+static gboolean
+inf_tcp_connection_send_real(InfTcpConnection* connection,
+                             gconstpointer data,
+                             guint* len)
+{
+  InfTcpConnectionPrivate* priv;
+  gconstpointer send_data;
+  guint send_len;
+  int errcode;
+  ssize_t result;
+
+  priv = INF_TCP_CONNECTION_PRIVATE(connection);
+  g_assert(priv->status == INF_TCP_CONNECTION_CONNECTED);
+
+  g_assert(data != NULL);
+  g_assert(len != NULL);
+
+  send_data = data;
+  send_len = *len;
+
+  do
+  {
+    result = send(
+      priv->socket,
+      send_data,
+      send_len,
+      INF_NATIVE_SOCKET_SENDRECV_FLAGS
+    );
+
+    /* Preserve error code so that it is not modified by future calls */
+    errcode = INF_NATIVE_SOCKET_LAST_ERROR;
+
+    if(result < 0 &&
+       errcode != INF_NATIVE_SOCKET_EINTR &&
+       errcode != INF_NATIVE_SOCKET_EAGAIN)
+    {
+      inf_tcp_connection_system_error(connection, errcode);
+      return FALSE;
+    }
+    else if(result == 0)
+    {
+      inf_tcp_connection_close(connection);
+      return FALSE;
+    }
+    else if(result > 0)
+    {
+      send_data = (const char*)send_data + result;
+      send_len -= result;
+    }
+  } while( (send_len > 0) &&
+           (result > 0 || errcode == INF_NATIVE_SOCKET_EINTR) &&
+           (priv->socket != INVALID_SOCKET) );
+
+  *len -= send_len;
+  return TRUE;
+}
+
 static void
 inf_tcp_connection_io_incoming(InfTcpConnection* connection)
 {
@@ -330,12 +599,12 @@ inf_tcp_connection_io_incoming(InfTcpConnection* connection)
 
   do
   {
-    result = recv(priv->socket, buf, 2048, INF_TCP_CONNECTION_SENDRECV_FLAGS);
-    errcode = INF_TCP_CONNECTION_LAST_ERROR;
+    result = recv(priv->socket, buf, 2048, INF_NATIVE_SOCKET_SENDRECV_FLAGS);
+    errcode = INF_NATIVE_SOCKET_LAST_ERROR;
 
     if(result < 0 &&
-       errcode != INF_TCP_CONNECTION_EINTR &&
-       errcode != INF_TCP_CONNECTION_EAGAIN)
+       errcode != INF_NATIVE_SOCKET_EINTR &&
+       errcode != INF_NATIVE_SOCKET_EAGAIN)
     {
       inf_tcp_connection_system_error(connection, errcode);
     }
@@ -354,7 +623,7 @@ inf_tcp_connection_io_incoming(InfTcpConnection* connection)
       );
     }
   } while( ((result > 0) ||
-            (result < 0 && errcode == INF_TCP_CONNECTION_EINTR)) &&
+            (result < 0 && errcode == INF_NATIVE_SOCKET_EINTR)) &&
            (priv->status != INF_TCP_CONNECTION_CLOSED));
 }
 
@@ -437,11 +706,14 @@ inf_tcp_connection_io(InfNativeSocket* socket,
   InfTcpConnectionPrivate* priv;
   socklen_t len;
   int errcode;
+  GError* error;
 
   connection = INF_TCP_CONNECTION(user_data);
   priv = INF_TCP_CONNECTION_PRIVATE(connection);
   g_object_ref(G_OBJECT(connection));
 
+  g_assert(priv->status != INF_TCP_CONNECTION_CLOSED);
+
   if(events & INF_IO_ERROR)
   {
     len = sizeof(int);
@@ -457,9 +729,32 @@ inf_tcp_connection_io(InfNativeSocket* socket,
     /* TODO: Maybe we should change this by mapping G_IO_HUP to
      * INF_IO_INCOMING, hoping recv() does the right thing then. */
     if(errcode != 0)
-      inf_tcp_connection_system_error(connection, errcode);
+    {
+      error = NULL;
+      inf_native_socket_make_error(errcode, &error);
+
+      if(priv->status == INF_TCP_CONNECTION_CONNECTING)
+      {
+        inf_tcp_connection_connection_error(connection, error);
+      }
+      else
+      {
+        g_signal_emit(
+          G_OBJECT(connection),
+          tcp_connection_signals[ERROR_],
+          0,
+          error
+        );
+      }
+
+      /* Error has been reported via signal emission, and there is nothing
+       * else to do with it. */
+      g_error_free(error);
+    }
     else
+    {
       inf_tcp_connection_close(connection);
+    }
   }
   else
   {
@@ -480,6 +775,70 @@ inf_tcp_connection_io(InfNativeSocket* socket,
   g_object_unref(G_OBJECT(connection));
 }
 
+static void
+inf_tcp_connection_resolved_cb(InfNameResolver* resolver,
+                               const GError* error,
+                               gpointer user_data)
+{
+  InfTcpConnection* connection;
+  InfTcpConnectionPrivate* priv;
+
+  connection = INF_TCP_CONNECTION(user_data);
+  priv = INF_TCP_CONNECTION_PRIVATE(connection);
+
+  if(priv->status == INF_TCP_CONNECTION_CONNECTING)
+  {
+    if(error != NULL)
+    {
+      /* If there was an error, no additional addresses are available */
+      g_assert(
+        priv->resolver_index == inf_name_resolver_get_n_addresses(resolver)
+      );
+
+      inf_tcp_connection_connection_error(connection, error);
+    }
+    else
+    {
+      /* If there was no error, try opening a connection to the resolved
+       * address(es). */
+      inf_tcp_connection_open_with_resolver(connection, NULL);
+    }
+  }
+}
+
+static void
+inf_tcp_connection_set_resolver(InfTcpConnection* connection,
+                                InfNameResolver* resolver)
+{
+  InfTcpConnectionPrivate* priv;
+  priv = INF_TCP_CONNECTION_PRIVATE(connection);
+
+  if(priv->resolver != NULL)
+  {
+    inf_signal_handlers_disconnect_by_func(
+      G_OBJECT(priv->resolver),
+      G_CALLBACK(inf_tcp_connection_resolved_cb),
+      connection
+    );
+
+    g_object_unref(priv->resolver);
+  }
+
+  priv->resolver = resolver;
+
+  if(resolver != NULL)
+  {
+    g_object_ref(resolver);
+
+    g_signal_connect(
+      G_OBJECT(resolver),
+      "resolved",
+      G_CALLBACK(inf_tcp_connection_resolved_cb),
+      connection
+    );
+  }
+}
+
 static void
 inf_tcp_connection_init(GTypeInstance* instance,
                         gpointer g_class)
@@ -493,6 +852,8 @@ inf_tcp_connection_init(GTypeInstance* instance,
   priv->io = NULL;
   priv->events = 0;
   priv->watch = NULL;
+  priv->resolver = NULL;
+  priv->resolver_index = 0;
   priv->status = INF_TCP_CONNECTION_CLOSED;
   priv->socket = INVALID_SOCKET;
 
@@ -518,9 +879,11 @@ inf_tcp_connection_dispose(GObject* object)
   if(priv->status != INF_TCP_CONNECTION_CLOSED)
     inf_tcp_connection_close(connection);
 
+  inf_tcp_connection_set_resolver(connection, NULL);
+
   if(priv->io != NULL)
   {
-    g_object_unref(G_OBJECT(priv->io));
+    g_object_unref(priv->io);
     priv->io = NULL;
   }
 
@@ -569,6 +932,15 @@ inf_tcp_connection_set_property(GObject* object,
     g_assert(priv->status == INF_TCP_CONNECTION_CLOSED);
     if(priv->io != NULL) g_object_unref(G_OBJECT(priv->io));
     priv->io = INF_IO(g_value_dup_object(value));
+    break;
+  case PROP_RESOLVER:
+    g_assert(priv->status == INF_TCP_CONNECTION_CLOSED);
+
+    inf_tcp_connection_set_resolver(
+      connection,
+      INF_NAME_RESOLVER(g_value_get_object(value))
+    );
+
     break;
   case PROP_REMOTE_ADDRESS:
     g_assert(priv->status == INF_TCP_CONNECTION_CLOSED);
@@ -636,6 +1008,9 @@ inf_tcp_connection_get_property(GObject* object,
   case PROP_IO:
     g_value_set_object(value, G_OBJECT(priv->io));
     break;
+  case PROP_RESOLVER:
+    g_value_set_object(value, G_OBJECT(priv->resolver));
+    break;
   case PROP_STATUS:
     g_value_set_enum(value, priv->status);
     break;
@@ -779,6 +1154,18 @@ inf_tcp_connection_class_init(gpointer g_class,
     )
   );
 
+  g_object_class_install_property(
+    object_class,
+    PROP_RESOLVER,
+    g_param_spec_object(
+      "resolver",
+      "Resolver",
+      "The hostname resolver",
+      INF_TYPE_NAME_RESOLVER,
+      G_PARAM_READWRITE | G_PARAM_CONSTRUCT
+    )
+  );
+
   g_object_class_install_property(
     object_class,
     PROP_STATUS,
@@ -872,9 +1259,11 @@ inf_tcp_connection_class_init(gpointer g_class,
 
   /**
    * InfTcpConnection::sent:
-   * @connection: The #InfTcpConnection through which the data has been sent
-   * @data: A #gpointer refering to the data that has been sent
-   * @length: A #guint holding the number of bytes that has been sent
+   * @connection: The #InfTcpConnection through which the data has been sent.
+   * @data: A #gpointer refering to the data that has been sent.
+   * @length: A #guint holding the number of bytes that has been sent.
+   *
+   * This signal is emitted whenever data has been sent over the connection.
    */
   tcp_connection_signals[SENT] = g_signal_new(
     "sent",
@@ -891,9 +1280,13 @@ inf_tcp_connection_class_init(gpointer g_class,
 
   /**
    * InfTcpConnection::received:
-   * @connection: The #InfTcpConnection through which the data has been received
-   * @data: A #gpointer refering to the data that has been received
-   * @length: A #guint holding the number of bytes that has been received
+   * @connection: The #InfTcpConnection through which the data has been
+   * received.
+   * @data: A #gpointer refering to the data that has been received.
+   * @length: A #guint holding the number of bytes that has been received.
+   *
+   * This signal is emitted whenever data has been received from the
+   * connection.
    */
   tcp_connection_signals[RECEIVED] = g_signal_new(
     "received",
@@ -910,8 +1303,12 @@ inf_tcp_connection_class_init(gpointer g_class,
 
   /**
    * InfTcpConnection::error:
-   * @connection: The erroneous #InfTcpConnection
-   * @error: A pointer to a #GError object with details on the error
+   * @connection: The erroneous #InfTcpConnection.
+   * @error: A pointer to a #GError object with details on the error.
+   *
+   * This signal is emitted when an error occured with the connection. If the
+   * error is fatal, the connection will change its status to
+   * %INF_TCP_CONNECTION_CLOSED.
    */
   tcp_connection_signals[ERROR_] = g_signal_new(
     "error",
@@ -1007,7 +1404,7 @@ inf_tcp_connection_get_type(void)
  **/
 InfTcpConnection*
 inf_tcp_connection_new(InfIo* io,
-                       InfIpAddress* remote_addr,
+                       const InfIpAddress* remote_addr,
                        guint remote_port)
 {
   InfTcpConnection* tcp;
@@ -1022,7 +1419,9 @@ inf_tcp_connection_new(InfIo* io,
       "io", io,
       "remote-address", remote_addr,
       "remote-port", remote_port,
-      NULL));
+      NULL
+    )
+  );
 
   return tcp;
 }
@@ -1035,14 +1434,14 @@ inf_tcp_connection_new(InfIo* io,
  * @error: Location to store error information.
  *
  * Creates a new #InfTcpConnection and connects it to the given TCP endpoint.
- * Like inf_tcp_connection_new, but calls inf_tcp_connection_open().
+ * Like inf_tcp_connection_new(), but calls inf_tcp_connection_open().
  *
  * Returns: A new #InfTcpConnection, or %NULL on error. Free with
  * g_object_unref().
  **/
 InfTcpConnection*
 inf_tcp_connection_new_and_open(InfIo* io,
-                                InfIpAddress* remote_addr,
+                                const InfIpAddress* remote_addr,
                                 guint remote_port,
                                 GError** error)
 {
@@ -1064,6 +1463,44 @@ inf_tcp_connection_new_and_open(InfIo* io,
   return tcp;
 }
 
+/**
+ * inf_tcp_connection_new_resolve:
+ * @io: A #InfIo object used to watch for activity.
+ * @resolver: The hostname resolver object used to look up the remote
+ * hostname.
+ *
+ * Creates a new #InfTcpConnection and instead of setting the remote IP
+ * address and port number directly, a hostname resolver is used to look up
+ * the remote hostname before connecting. This has the advantage that all
+ * available addresses for that hostname are tried before giving up.
+ *
+ * The argument is stored as a property for an eventual
+ * inf_tcp_connection_open() call, this function itself does not
+ * establish a connection.
+ *
+ * Returns: A new #InfTcpConnection. Free with g_object_unref().
+ */
+InfTcpConnection*
+inf_tcp_connection_new_resolve(InfIo* io,
+                               InfNameResolver* resolver)
+{
+  InfTcpConnection* tcp;
+
+  g_return_val_if_fail(INF_IS_IO(io), NULL);
+  g_return_val_if_fail(INF_IS_NAME_RESOLVER(resolver), NULL);
+
+  tcp = INF_TCP_CONNECTION(
+    g_object_new(
+      INF_TYPE_TCP_CONNECTION,
+      "io", io,
+      "resolver", resolver,
+      NULL
+    )
+  );
+
+  return tcp;
+}
+
 /**
  * inf_tcp_connection_open:
  * @connection: A #InfTcpConnection.
@@ -1084,162 +1521,38 @@ inf_tcp_connection_open(InfTcpConnection* connection,
                         GError** error)
 {
   InfTcpConnectionPrivate* priv;
-/*  char device_name[IF_NAMESIZE];*/
-
-#ifdef G_OS_WIN32
-  u_long argp;
-#endif
-
-  union {
-    struct sockaddr_in in;
-    struct sockaddr_in6 in6;
-  } native_address;
-
-  struct sockaddr* addr;
-  socklen_t addrlen;
-  int result;
-  int errcode;
 
   g_return_val_if_fail(INF_IS_TCP_CONNECTION(connection), FALSE);
   priv = INF_TCP_CONNECTION_PRIVATE(connection);
 
   g_return_val_if_fail(priv->io != NULL, FALSE);
   g_return_val_if_fail(priv->status == INF_TCP_CONNECTION_CLOSED, FALSE);
-  g_return_val_if_fail(priv->remote_address != NULL, FALSE);
-  g_return_val_if_fail(priv->remote_port != 0, FALSE);
 
-  /* Close previous socket */
-  if(priv->socket != INVALID_SOCKET)
-    closesocket(priv->socket);
+  g_return_val_if_fail(
+    priv->remote_address != NULL || priv->resolver != NULL,
+    FALSE
+  );
 
-  switch(inf_ip_address_get_family(priv->remote_address))
+  g_return_val_if_fail(
+    priv->remote_port != 0 ||
+    priv->resolver != NULL,
+    FALSE
+  );
+
+  if(priv->resolver != NULL)
   {
-  case INF_IP_ADDRESS_IPV4:
-    priv->socket = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
-    addr = (struct sockaddr*)&native_address.in;
-    addrlen = sizeof(struct sockaddr_in);
-
-    memcpy(
-      &native_address.in.sin_addr,
-      inf_ip_address_get_raw(priv->remote_address),
-      sizeof(struct in_addr)
-    );
-
-    native_address.in.sin_family = AF_INET;
-    native_address.in.sin_port = htons(priv->remote_port);
-
-    break;
-  case INF_IP_ADDRESS_IPV6:
-    priv->socket = socket(PF_INET6, SOCK_STREAM, IPPROTO_TCP);
-    addr = (struct sockaddr*)&native_address.in6;
-    addrlen = sizeof(struct sockaddr_in6);
-
-    memcpy(
-      &native_address.in6.sin6_addr,
-      inf_ip_address_get_raw(priv->remote_address),
-      sizeof(struct in6_addr)
-    );
-
-    native_address.in6.sin6_family = AF_INET6;
-    native_address.in6.sin6_port = htons(priv->remote_port);
-    native_address.in6.sin6_flowinfo = 0;
-    native_address.in6.sin6_scope_id = priv->device_index;
-
-    break;
-  default:
-    g_assert_not_reached();
-    break;
-  }
-
-  if(priv->socket == INVALID_SOCKET)
-  {
-    inf_tcp_connection_make_system_error(
-      INF_TCP_CONNECTION_LAST_ERROR,
-      error
-    );
-
-    return FALSE;
-  }
-
-  /* Set socket non-blocking */
-#ifndef G_OS_WIN32
-  result = fcntl(priv->socket, F_GETFL);
-  if(result == INVALID_SOCKET)
-  {
-    errcode = INF_TCP_CONNECTION_LAST_ERROR;
-    inf_tcp_connection_make_system_error(errcode, error);
-
-    closesocket(priv->socket);
-    priv->socket = INVALID_SOCKET;
-    return FALSE;
-  }
-
-  if(fcntl(priv->socket, F_SETFL, result | O_NONBLOCK) == -1)
-  {
-    errcode = INF_TCP_CONNECTION_LAST_ERROR;
-    inf_tcp_connection_make_system_error(errcode, error);
-
-    closesocket(priv->socket);
-    priv->socket = INVALID_SOCKET;
-    return FALSE;
-  }
-#else
-  argp = 1;
-  if(ioctlsocket(priv->socket, FIONBIO, &argp) != 0)
-  {
-    errcode = INF_TCP_CONNECTION_LAST_ERROR;
-    inf_tcp_connection_make_system_error(errcode, error);
-
-    closesocket(priv->socket);
-    priv->socket = INVALID_SOCKET;
-    return FALSE;
-  }
-#endif
-
-  /* Connect */
-  do
-  {
-    result = connect(priv->socket, addr, addrlen);
-    errcode = INF_TCP_CONNECTION_LAST_ERROR;
-    if(result == -1 &&
-       errcode != INF_TCP_CONNECTION_EINTR &&
-       errcode != INF_TCP_CONNECTION_EINPROGRESS)
-    {
-      inf_tcp_connection_make_system_error(errcode, error);
-
-      closesocket(priv->socket);
-      priv->socket = INVALID_SOCKET;
-
-      return FALSE;
-    }
-  } while(result == -1 && errcode != INF_TCP_CONNECTION_EINPROGRESS);
-
-  if(result == 0)
-  {
-    /* Connection fully established */
-    inf_tcp_connection_connected(connection);
+    g_assert(priv->resolver_index == 0);
+    return inf_tcp_connection_open_with_resolver(connection, error);
   }
   else
   {
-    g_assert(priv->watch == NULL);
-
-    /* Connection establishment in progress */
-    priv->events = INF_IO_OUTGOING | INF_IO_ERROR;
-
-    priv->watch = inf_io_add_watch(
-      priv->io,
-      &priv->socket,
-      priv->events,
-      inf_tcp_connection_io,
+    return inf_tcp_connection_open_real(
       connection,
-      NULL
+      priv->remote_address,
+      priv->remote_port,
+      error
     );
-
-    priv->status = INF_TCP_CONNECTION_CONNECTING;
-    g_object_notify(G_OBJECT(connection), "status");
   }
-
-  return TRUE;
 }
 
 /**
@@ -1444,23 +1757,23 @@ _inf_tcp_connection_accepted(InfIo* io,
   result = fcntl(socket, F_GETFL);
   if(result == -1)
   {
-    errcode = INF_TCP_CONNECTION_LAST_ERROR;
-    inf_tcp_connection_make_system_error(errcode, error);
+    errcode = INF_NATIVE_SOCKET_LAST_ERROR;
+    inf_native_socket_make_error(errcode, error);
     return NULL;
   }
 
   if(fcntl(socket, F_SETFL, result | O_NONBLOCK) == -1)
   {
-    errcode = INF_TCP_CONNECTION_LAST_ERROR;
-    inf_tcp_connection_make_system_error(errcode, error);
+    errcode = INF_NATIVE_SOCKET_LAST_ERROR;
+    inf_native_socket_make_error(errcode, error);
     return NULL;
   }
 #else
   argp = 1;
   if(ioctlsocket(socket, FIONBIO, &argp) != 0)
   {
-    errcode = INF_TCP_CONNECTION_LAST_ERROR;
-    inf_tcp_connection_make_system_error(errcode, error);
+    errcode = INF_NATIVE_SOCKET_LAST_ERROR;
+    inf_native_socket_make_error(errcode, error);
     return NULL;
   }
 #endif
