@@ -1,5 +1,5 @@
 /* libinfinity - a GObject-based infinote implementation
- * Copyright (C) 2007-2011 Armin Burgmeier <armin@arbur.net>
+ * Copyright (C) 2007-2014 Armin Burgmeier <armin@arbur.net>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -51,8 +51,6 @@ typedef struct _InfTextSessionPrivate InfTextSessionPrivate;
 struct _InfTextSessionPrivate {
   guint caret_update_interval;
   GSList* local_users;
-
-  gboolean apply_request;
 };
 
 enum {
@@ -87,6 +85,15 @@ static GQuark inf_text_session_error_quark;
  * Utility functions
  */
 
+static gint64
+inf_text_session_get_real_time()
+{
+  /* TODO: Replace by g_get_real_time() once we depend on glib >=2.28 */
+  GTimeVal timeval;
+  g_get_current_time(&timeval);
+  return (gint64)timeval.tv_sec * 1000000 + timeval.tv_usec;
+}
+
 /* Returns the difference between two GTimeVal, in milliseconds */
 static guint
 inf_text_session_timeval_diff(GTimeVal* first,
@@ -202,23 +209,51 @@ inf_text_session_broadcast_caret_selection(InfTextSession* session,
   InfAdoptedOperation* operation;
   InfAdoptedAlgorithm* algorithm;
   InfAdoptedRequest* request;
+  guint buf_len;
   guint position;
   int sel;
+  guint end;
 
   algorithm = inf_adopted_session_get_algorithm(INF_ADOPTED_SESSION(session));
   position = inf_text_user_get_caret_position(local->user);
   sel = inf_text_user_get_selection_length(local->user);
+  end = position + sel;
+
+  /* Clamp position and selection to buffer length. The only case when this is
+   * needed is when a local user's position is beyond the end of the document
+   * since there are some local document modifications. This can happen with
+   * for example InfTextFixlineBuffer. */
+  /* TODO: This should be handled more cleverly, by propagating the user
+   * position and selection through the buffer, to make sure that at this
+   * point it is always consistent with the infinote view of the buffer. */
+  buf_len = inf_text_buffer_get_length(
+    INF_TEXT_BUFFER(inf_session_get_buffer(INF_SESSION(session)))
+  );
+
+  if(position > buf_len)
+    position = buf_len;
+  if(end > buf_len)
+    end = buf_len;
+
+  if(end >= position)
+    sel = (int)(end - position);
+  else
+    sel = -(int)(position - end);
 
   operation = INF_ADOPTED_OPERATION(
     inf_text_move_operation_new(position, sel)
   );
 
-  request = inf_adopted_algorithm_generate_request_noexec(
+  request = inf_adopted_algorithm_generate_request(
     algorithm,
+    INF_ADOPTED_REQUEST_DO,
     INF_ADOPTED_USER(local->user),
     operation
   );
 
+  /* This cannot fail since operation is not applied */
+  inf_adopted_algorithm_execute_request(algorithm, request, FALSE, NULL);
+
   g_object_unref(operation);
 
   inf_adopted_session_broadcast_request(
@@ -260,20 +295,23 @@ inf_text_session_selection_changed_cb(InfTextUser* user,
 {
   InfTextSession* session;
   InfTextSessionPrivate* priv;
+  InfAdoptedAlgorithm* algorithm;
+  InfAdoptedRequest* execute_request;
   InfTextSessionLocalUser* local;
   GTimeVal current;
   guint diff;
 
   session = INF_TEXT_SESSION(user_data);
   priv = INF_TEXT_SESSION_PRIVATE(session);
-
+  algorithm = inf_adopted_session_get_algorithm(INF_ADOPTED_SESSION(session));
+  execute_request = inf_adopted_algorithm_get_execute_request(algorithm);
 
   /* We should block all changes that have by_request set to FALSE... breaks
    * if someone else does that... should maybe emit a warning instead. */
-  g_assert( (priv->apply_request == TRUE && by_request == FALSE) ||
-            (priv->apply_request == FALSE && by_request == TRUE));
+  g_assert( (execute_request != NULL && by_request == FALSE) ||
+            (execute_request == NULL && by_request == TRUE));
 
-  if(priv->apply_request == FALSE)
+  if(execute_request == NULL)
   {
     local = inf_text_session_find_local_user(session, user);
     g_assert(local != NULL);
@@ -501,10 +539,12 @@ inf_text_session_buffer_text_inserted_cb(InfTextBuffer* buffer,
 {
   InfTextSession* session;
   InfTextSessionPrivate* priv;
-  InfAdoptedOperation* operation;
-  InfAdoptedAlgorithm* algorithm;
-  InfAdoptedRequest* request;
   InfUserTable* user_table;
+  InfAdoptedAlgorithm* algorithm;
+  InfAdoptedRequest* execute_request;
+
+  InfAdoptedOperation* operation;
+  InfAdoptedRequest* request;
   InfTextSessionInsertForeachData data;
 
   g_assert(INF_TEXT_IS_USER(user));
@@ -512,23 +552,25 @@ inf_text_session_buffer_text_inserted_cb(InfTextBuffer* buffer,
   session = INF_TEXT_SESSION(user_data);
   priv = INF_TEXT_SESSION_PRIVATE(session);
   user_table = inf_session_get_user_table(INF_SESSION(session));
+  algorithm = inf_adopted_session_get_algorithm(INF_ADOPTED_SESSION(session));
+  execute_request = inf_adopted_algorithm_get_execute_request(algorithm);
 
-  if(priv->apply_request == FALSE)
+  if(execute_request == NULL)
   {
     operation = INF_ADOPTED_OPERATION(
       inf_text_default_insert_operation_new(pos, chunk)
     );
 
-    algorithm = inf_adopted_session_get_algorithm(
-      INF_ADOPTED_SESSION(session)
-    );
-
-    request = inf_adopted_algorithm_generate_request_noexec(
+    request = inf_adopted_algorithm_generate_request(
       algorithm,
+      INF_ADOPTED_REQUEST_DO,
       INF_ADOPTED_USER(user),
       operation
     );
 
+    /* This cannot fail since operation is not applied */
+    inf_adopted_algorithm_execute_request(algorithm, request, FALSE, NULL);
+
     inf_adopted_session_broadcast_request(
       INF_ADOPTED_SESSION(session),
       request
@@ -575,10 +617,12 @@ inf_text_session_buffer_text_erased_cb(InfTextBuffer* buffer,
 {
   InfTextSession* session;
   InfTextSessionPrivate* priv;
-  InfAdoptedOperation* operation;
-  InfAdoptedAlgorithm* algorithm;
-  InfAdoptedRequest* request;
   InfUserTable* user_table;
+  InfAdoptedAlgorithm* algorithm;
+  InfAdoptedRequest* execute_request;
+
+  InfAdoptedOperation* operation;
+  InfAdoptedRequest* request;
   InfTextSessionEraseForeachData data;
 
   g_assert(INF_TEXT_IS_USER(user));
@@ -586,23 +630,25 @@ inf_text_session_buffer_text_erased_cb(InfTextBuffer* buffer,
   session = INF_TEXT_SESSION(user_data);
   priv = INF_TEXT_SESSION_PRIVATE(session);
   user_table = inf_session_get_user_table(INF_SESSION(session));
+  algorithm = inf_adopted_session_get_algorithm(INF_ADOPTED_SESSION(session));
+  execute_request = inf_adopted_algorithm_get_execute_request(algorithm);
 
-  if(priv->apply_request == FALSE)
+  if(execute_request == NULL)
   {
     operation = INF_ADOPTED_OPERATION(
       inf_text_default_delete_operation_new(pos, chunk)
     );
 
-    algorithm = inf_adopted_session_get_algorithm(
-      INF_ADOPTED_SESSION(session)
-    );
-
-    request = inf_adopted_algorithm_generate_request_noexec(
+    request = inf_adopted_algorithm_generate_request(
       algorithm,
+      INF_ADOPTED_REQUEST_DO,
       INF_ADOPTED_USER(user),
       operation
     );
 
+    /* This cannot fail since operation is not applied */
+    inf_adopted_algorithm_execute_request(algorithm, request, FALSE, NULL);
+
     inf_adopted_session_broadcast_request(
       INF_ADOPTED_SESSION(session),
       request
@@ -633,44 +679,6 @@ inf_text_session_buffer_text_erased_cb(InfTextBuffer* buffer,
   inf_text_session_unblock_local_users_selection_changed(session);
 }
 
-/* Remember when a buffer operation happens in response to the algorithm
- * applying a request so we don't re-broadcast incoming requests, and so that
- * we don't broadcast the effect of an Undo if the user calls
- * inf_adopted_session_undo(). */
-static void
-inf_text_session_apply_request_cb_before(InfAdoptedAlgorithm* algorithm,
-                                         InfAdoptedUser* user,
-                                         InfAdoptedRequest* request,
-                                         gpointer user_data)
-{
-  InfTextSession* session;
-  InfTextSessionPrivate* priv;
-  /*InfTextBuffer* buffer;*/
-
-  session = INF_TEXT_SESSION(user_data);
-  priv = INF_TEXT_SESSION_PRIVATE(session);
-  /*buffer = INF_TEXT_BUFFER(inf_session_get_buffer(INF_SESSION(session)));*/
-
-  priv->apply_request = TRUE;
-}
-
-static void
-inf_text_session_apply_request_cb_after(InfAdoptedAlgorithm* algorithm,
-                                        InfAdoptedUser* user,
-                                        InfAdoptedRequest* request,
-                                        gpointer user_data)
-{
-  InfTextSession* session;
-  InfTextSessionPrivate* priv;
-  /*InfTextBuffer* buffer;*/
-  
-  session = INF_TEXT_SESSION(user_data);
-  priv = INF_TEXT_SESSION_PRIVATE(session);
-  /*buffer = INF_TEXT_BUFFER(inf_session_get_buffer(INF_SESSION(session)));*/
-
-  priv->apply_request = FALSE;
-}
-
 static void
 inf_text_session_init_text_handlers_user_foreach_func(InfUser* user,
                                                       gpointer user_data)
@@ -722,20 +730,6 @@ inf_text_session_init_text_handlers(InfTextSession* session)
     session
   );
 
-  g_signal_connect(
-    G_OBJECT(algorithm),
-    "apply-request",
-    G_CALLBACK(inf_text_session_apply_request_cb_before),
-    session
-  );
-
-  g_signal_connect_after(
-    G_OBJECT(algorithm),
-    "apply-request",
-    G_CALLBACK(inf_text_session_apply_request_cb_after),
-    session
-  );
-
   inf_user_table_foreach_local_user(
     user_table,
     inf_text_session_init_text_handlers_user_foreach_func,
@@ -758,7 +752,6 @@ inf_text_session_init(GTypeInstance* instance,
   priv = INF_TEXT_SESSION_PRIVATE(session);
 
   priv->caret_update_interval = 500;
-  priv->apply_request = FALSE;
 }
 
 static GObject*
@@ -866,22 +859,6 @@ inf_text_session_dispose(GObject* object)
     session
   );
 
-  /* Algorithm may be NULL if we were still synchronizing */
-  if(algorithm != NULL)
-  {
-    inf_signal_handlers_disconnect_by_func(
-      G_OBJECT(algorithm),
-      G_CALLBACK(inf_text_session_apply_request_cb_before),
-      session
-    );
-
-    inf_signal_handlers_disconnect_by_func(
-      G_OBJECT(algorithm),
-      G_CALLBACK(inf_text_session_apply_request_cb_after),
-      session
-    );
-  }
-
   G_OBJECT_CLASS(parent_class)->dispose(object);
 }
 
@@ -1039,7 +1016,7 @@ inf_text_session_to_xml_sync(InfSession* session,
   buffer = INF_TEXT_BUFFER(inf_session_get_buffer(session));
   cd = g_iconv_open("UTF-8", inf_text_buffer_get_encoding(buffer));
 
-  iter = inf_text_buffer_create_iter(buffer);
+  iter = inf_text_buffer_create_begin_iter(buffer);
   if(iter != NULL)
   {
     result = TRUE;
@@ -1418,7 +1395,7 @@ inf_text_session_request_to_xml(InfAdoptedSession* session,
         INF_TEXT_DEFAULT_INSERT_OPERATION(operation)
       );
 
-      result = inf_text_chunk_iter_init(chunk, &iter);
+      result = inf_text_chunk_iter_init_begin(chunk, &iter);
       g_assert(result == TRUE);
 
       utf8_text = g_convert(
@@ -1465,7 +1442,7 @@ inf_text_session_request_to_xml(InfAdoptedSession* session,
 
         /* Need to transmit all deleted data */
         cd = g_iconv_open("UTF-8", inf_text_chunk_get_encoding(chunk));
-        result = inf_text_chunk_iter_init(chunk, &iter);
+        result = inf_text_chunk_iter_init_begin(chunk, &iter);
 
         while(result == TRUE)
         {
@@ -1764,14 +1741,27 @@ inf_text_session_xml_to_request(InfAdoptedSession* session,
   {
   case INF_ADOPTED_REQUEST_DO:
     g_assert(operation != NULL);
-    request = inf_adopted_request_new_do(vector, user_id, operation);
+    request = inf_adopted_request_new_do(
+      vector,
+      user_id,
+      operation,
+      inf_text_session_get_real_time()
+    );
     g_object_unref(operation);
     break;
   case INF_ADOPTED_REQUEST_UNDO:
-    request = inf_adopted_request_new_undo(vector, user_id);
+    request = inf_adopted_request_new_undo(
+      vector,
+      user_id,
+      inf_text_session_get_real_time()
+    );
     break;
   case INF_ADOPTED_REQUEST_REDO:
-    request = inf_adopted_request_new_redo(vector, user_id);
+    request = inf_adopted_request_new_redo(
+      vector,
+      user_id,
+      inf_text_session_get_real_time()
+    );
     break;
   default:
     g_assert_not_reached();
@@ -2083,4 +2073,89 @@ inf_text_session_flush_requests_for_user(InfTextSession* session,
   }
 }
 
+/**
+ * inf_text_session_join_user:
+ * @proxy: A #InfSessionProxy with a #InfTextSession session.
+ * @name: The name of the user to join.
+ * @status: The initial status of the user to join. Must not be
+ * @INF_USER_UNAVAILABLE.
+ * @hue: The user color of the user to join.
+ * @caret_position: The initial position of the new user's cursor.
+ * @selection_length: The initial length of the new user's selection.
+ * @func: Function to call after completion of the request, or %NULL.
+ * @user_data: Additional data to pass to @func.
+ *
+ * This functions creates a user join request for an #InfTextSession. This is
+ * a shortcut for inf_session_proxy_join_user().
+ *
+ * Returns: A #InfRequest, or %NULL on error.
+ */
+InfRequest*
+inf_text_session_join_user(InfSessionProxy* proxy,
+                           const gchar* name,
+                           InfUserStatus status,
+                           gdouble hue,
+                           guint caret_position,
+                           int selection_length,
+                           InfRequestFunc func,
+                           gpointer user_data)
+{
+#define N_PARAMS 6u
+  GParameter params[N_PARAMS] = {
+    { "hue", { 0 } },
+    { "vector", { 0 } },
+    { "caret-position", { 0 } },
+    { "selection-length", { 0 } },
+    { "name", { 0 } },
+    { "status", { 0 } }
+  };
+
+  InfSession* session;
+  InfRequest* request;
+  guint i;
+
+  g_return_val_if_fail(INF_IS_SESSION_PROXY(proxy), NULL);
+  
+  g_object_get(G_OBJECT(proxy), "session", &session, NULL);
+  g_return_val_if_fail(INF_TEXT_IS_SESSION(session), NULL);
+
+  g_value_init(&params[0].value, G_TYPE_DOUBLE);
+  g_value_set_double(&params[0].value, hue);
+
+  g_value_init(&params[1].value, INF_ADOPTED_TYPE_STATE_VECTOR);
+  g_value_set_boxed(
+    &params[1].value,
+    inf_adopted_algorithm_get_current(
+      inf_adopted_session_get_algorithm(INF_ADOPTED_SESSION(session))
+    )
+  );
+
+  g_value_init(&params[2].value, G_TYPE_UINT);
+  g_value_set_uint(&params[2].value, caret_position);
+
+  g_value_init(&params[3].value, G_TYPE_INT);
+  g_value_set_int(&params[3].value, selection_length);
+
+  g_value_init(&params[4].value, G_TYPE_STRING);
+  g_value_set_string(&params[4].value, name); /* TODO: set_static_string? */
+
+  g_value_init(&params[5].value, INF_TYPE_USER_STATUS);
+  g_value_set_enum(&params[5].value, status);
+
+  request = inf_session_proxy_join_user(
+    proxy,
+    N_PARAMS,
+    params,
+    func,
+    user_data
+  );
+
+  for(i = 0; i < N_PARAMS; ++i)
+    g_value_unset(&params[i].value);
+#undef N_PARAMS
+
+  g_object_unref(session);
+  return request;
+}
+
 /* vim:set et sw=2 ts=2: */
